package ecs.exam2018;

import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.util.Set;
import java.util.HashSet;
/*@ init @*/
class TCPRegistryConnection extends Thread {
  private PrintStream out;
  private BufferedReader in;
  private Socket toClient;
  private Set<String> registry;

  TCPRegistryConnection(Socket s, Set<String> r)
    throws IOException {
//  (*\circled{1}*)  /*@ before1 @*/
      toClient = s;
      out = new PrintStream(toClient.getOutputStream());
      in = new BufferedReader(new InputStreamReader(toClient.getInputStream()));
      registry = r;
      /*@ after1 @*/
  }

  boolean add(String ip) {
//   (*\circled{2}*)  /*@ before2 @*/
    synchronized (registry) {
      return registry.add(ip);
    }
    /*@ after2 @*/
  }

  boolean remove(String ip) {
//   (*\circled{3}*)  /*@ before3 @*/
    synchronized (registry) {
      return registry.remove(ip);
    }
    /*@ after3 @*/
  }

  String getClientIP() {
//    (*\circled{4}*)   /*@ before4 @*/
    return toClient.getInetAddress().toString();
    /*@ after4 @*/
  }

  public void run() {
    try {
//   (*\circled{5}*)   /*@ before5 @*/
      String msg;
      switch (in.readLine()) {

      case "REGISTER":
        msg = add(getClientIP()) ? "OK" : "ERROR";
        out.println(msg);
        break;
      case "UNREGISTER":
        msg = remove(getClientIP()) ? "OK" : "ERROR";
        out.println(msg);
        break;
      case "LIST":
        synchronized (registry) {
          out.println("BEGIN");
          for (String ip : registry) {
            out.println(ip);
          }
          out.println("END");
        }
      default:
        out.println("ERROR");
      }
      toClient.close();
      /*@ after5 @*/
    } catch (IOException e) { }
  }
}
/*@ beforeclass @*/
public class TCPRegistryServer {

  private ServerSocket serverSocket;
  private Set<String> registry;

  public TCPRegistryServer() {
    registry = new HashSet<String>();
  };

  public void start(int port) throws IOException {
    if (serverSocket == null) {
      serverSocket = new ServerSocket();
    }
    serverSocket.bind(new InetSocketAddress(port));
  }
  public void mainLoop() {
    while (true) {
      try {
        Socket s = serverSocket.accept();
        TCPRegistryConnection c =
          new TCPRegistryConnection(s, registry);
        c.start();
      } catch (IOException e) {
        return;
      }
    }
  }
/*@ afterclass @*/
  public static void main(String[] args) {

    try {
      TCPRegistryServer s = new TCPRegistryServer();
      s.start(12345);
      s.mainLoop();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

}
